/*
 * Copyright (c) 2011 Marc R. Hoffmann, hoffmann@mountainminds.com
 * 
 * This work is licensed under the Creative Commons CC BY-SA License. To view a
 * copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/
 */
package com.mountainminds.examples.java7.jvm;

import java.lang.reflect.Modifier;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodHandle;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.GeneratorAdapter;
import org.objectweb.asm.commons.Method;

/**
 * Factory to generate a classes that contain invokedynamic bytecode
 * instructions which otherwise are not emitted by the Java compiler. Using the
 * ASM library (Version 4.0 RC 1) the factory creates subclasses of abstract
 * classes. All abstract methods are implemented with a single call site using
 * the same signature than the abstract method declaration.
 */
public class InvokeDynamicFactory {

	private static final Type GENERATED_TYPE = Type
			.getObjectType("com/mountainminds/examples/java7/jvm/InvokeDynamicFactory$Generated");

	private final ClassLoader parent;

	public InvokeDynamicFactory(ClassLoader parent) {
		this.parent = parent;
	}

	public <T> T create(Class<T> basetype, String bootstrapType,
			String bootstrapName, String bootstrapDesc) throws Exception {
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS
				| ClassWriter.COMPUTE_FRAMES);
		generateClass(cw, basetype, bootstrapType, bootstrapName, bootstrapDesc);
		return basetype.cast(MemoryClassLoader.load(parent,
				GENERATED_TYPE.getClassName(), cw.toByteArray()).newInstance());
	}

	private void generateClass(ClassVisitor cv, Class<?> baseclass,
			String bootstrapType, String bootstrapName, String bootstrapDesc) {
		Type basetype = Type.getType(baseclass);
		cv.visit(Opcodes.V1_7, Opcodes.ACC_PUBLIC,
				GENERATED_TYPE.getInternalName(), null,
				basetype.getInternalName(), null);

		generateInit(cv, basetype);

		for (java.lang.reflect.Method m : baseclass.getMethods()) {
			if (Modifier.isAbstract(m.getModifiers())) {
				generateDelegateMethod(Method.getMethod(m), cv, bootstrapType,
						bootstrapName, bootstrapDesc);
			}
		}

		cv.visitEnd();
	}

	private void generateInit(ClassVisitor cv, Type superclass) {
		Method method = Method.getMethod("void <init> ()");
		GeneratorAdapter mg = new GeneratorAdapter(Opcodes.ACC_PUBLIC, method,
				null, null, cv);
		mg.loadThis();
		mg.invokeConstructor(superclass, method);
		mg.returnValue();
		mg.endMethod();
	}

	private void generateDelegateMethod(Method method, ClassVisitor cv,
			String bootstrapType, String bootstrapName, String bootstrapDesc) {
		GeneratorAdapter mg = new GeneratorAdapter(Opcodes.ACC_PUBLIC, method,
				null, null, cv);
		mg.loadArgs();
		mg.visitInvokeDynamicInsn(method.getName(), method.getDescriptor(),
				new MethodHandle(Opcodes.MH_INVOKESTATIC, bootstrapType,
						bootstrapName, bootstrapDesc));
		mg.returnValue();
		mg.endMethod();
	}

}
